'******************************************************************************
'*  io.spin - handle input and output requests sent by pm80
'*
'*  (c) Juergen Buchmueller <pullmoll@t-online.de>
'*
'* $Id: io.spin,v 1.20 2010-03-25 09:40:35 pm Exp $
'******************************************************************************
CON

	io_cmd_out		=	$01		' I/O emulator command output
	io_cmd_in		=	$02		' I/O emulator command input
	io_cmd_break		=	$03		' I/O emulator command break
	io_cmd_kbd		=	$04		' I/O emulator command read TRS-80 keyboard

	baud			=	115200		' baud rate

VAR
	long	cog
	long	io_command
	long	params[2]
	long	rc
	word	topy
	word	boty
	word	row
	word	y0
	byte	curx
	byte	cury
	byte	de[32]					' directory entry
	byte	name[12+1]
	byte	datetime[20+1]
	byte	buff[32]

OBJ
	fat	: "fatfs"
	video	: "trs80_video"
	kbd	: "trs80_keyboard"
#ifdef TriBladeProp
	xmm	: "TriBladeProp_Blade2_204"		' Extended memory manager
#endif
#ifdef RamBladeProp
	xmm	: "RamBlade_203_002"			' Extended memory manager
#endif
#ifdef DracBladeProp
	xmm	: "Dracblade"				' Extended memory manager
#endif

PUB start(SI, SO, doPin, clkPin, diPin, csPin, kbDpin, kbCpin, auPin, tvPin, screen, iobuff) : okay | i, dent, sec
	stop

	screen_ptr := screen
	iobuff_ptr := iobuff
	chr(12)
	str(string("********* TRS-80 emulator on the Propeller *********",13,10))

	okay := \fat.start(iobuff_ptr, doPin, clkPin, diPin, csPin)
	if okay < 0
		failure(string("FAT + sdspi"),okay)
		return -1
	success(string("FAT + sdspi"))

	params[0] := screen				' 1K video RAM
	params[1] := tvPin				' video pin
	video.start(@params)

	okay := \kbd.startx(kbDpin, kbCpin, 0, 40)
	if okay < 0
		failure(string("Keyboard"),okay)
		return
	success(string("Keyboard"))
	kbd_states := kbd.states_ptr

	dira_mask := 1 << auPin
	ctra_mode := %00110 << 26 | auPin		' set to single ended DUTY mode

	okay := cog := COGNEW(@entry, @io_command) + 1
	if okay =< 0
		failure(string("I/O"),okay)
		return -1
	success(string("I/O"))

	okay := \xmm.start
	if okay < 0
		failure(string("XMM"), okay)
		return -1
	success(string("XMM"))
	xmm_command := xmm.rendezvous
	xmm_hubaddr := xmm_command + 1*4
	xmm_ramaddr := xmm_command + 2*4
	xmm_blocklen := xmm_command + 3*4
	xmm_errx := xmm_command + 4*4

	' clear the Z80 RAM
	longfill(iobuff_ptr, 0, 512/4)
	repeat i from $4000 to $fe00 step $200
		xmm.DoCmd("W", iobuff_ptr, i, $200)

	okay := \fat.mount
	if okay < 0
		failure(string("FAT mount"), okay)
		return -1
	str(string("FAT"))
	dec(fat.get_type)
	str(string(" directory entries: "))
	dec(fat.get_dir_size)
	crlf

	dent := open_file(string("TRS80.ROM"))
	if dent < 0
		return -1
	sec := fat.get_dent_first_sector(@de)
	repeat i from $0000 to $2e00 step $200
		fat.read_sector(sec)
		xmm.DoCmd("W", iobuff_ptr, i, $200)
		sec := fat.get_dent_next_sector(@de, sec)

	row := 0
	topy := 0
	str(string("Select file to load to memory: <UP> <DN> <RETURN>"))
	repeat
		boty := 9999
		repeat y0 from 0 to 6
			rc := direntry(topy + y0, @buff)
			if rc < 0
				bytefill(@buff, " ", 14)
				buff[14] := 0
				if boty == 9999
					boty := y0
			gotoxy(0, 8+y0)
			if y0 == row
				str(string("==> "))
			else
				str(string("    "))
			str(@buff)
			if y0 == row
				str(string(" <==          "))
			else
				str(string("              "))
		if kbd.keystate($34)	' down arrow
			if row < 6
				if row < boty
					row++
			else
				if boty == 9999
					topy++
		if kbd.keystate($33)
			if row > 0
				row--
			elseif topy > 0
				topy--
		if kbd.keystate($30)
			bytefill(@buff, " ", 32)
			buff[31] := 0
			repeat y0 from 0 to 6
				gotoxy(0, 8+y0)
				str(@buff)
			gotoxy(0, 8)
			rc := direntry(topy + row, @buff)
			quit
		
	cmd_load(@buff)

	str(string("Press <RETURN> to start...",13,10))
	repeat until !kbd.keystate($30)
	repeat until kbd.keystate($30)

	return 0

PUB stop
'' Stop CPU simulator - frees a cog
	if cog
		cogstop(cog~ - 1)

PRI gotoxy(x,y)
	curx := (x #> 0) <# 63
	cury := (y #> 0) <# 15

PRI chr(ch)
	case ch
		13:
			curx := 0
		12:
			longfill(screen_ptr, $20202020, 64*16/4)
			curx := 0
			cury := 0
		10:
			cury += 1
			if cury > 15
				longmove(screen_ptr, screen_ptr + 64, 64*15/4)
				longfill(screen_ptr + 64*15, $20202020, 64/4)
				cury := 15
		9:
			repeat until (curx & 7) == 0
				chr(32)
		8:
			if curx > 0
				--curx
		other:
			byte[screen_ptr][cury << 6 + curx] := ch
			curx++
			if curx == 64
				chr(13)
				chr(10)

PRI str(str_ptr) | i
	repeat i from 0 to strsize(str_ptr) - 1
		chr(byte[str_ptr][i])

PRI crlf
	chr(13)
	chr(10)

PRI dec_size(value)
	if value > 1048576
		dec(value/1048576)
		chr(".")
		dec(10*(value // 1048576) / 1048576)
		str(string("MB"))
	elseif value > 1024
		dec(value/1024)
		chr(".")
		dec(10*(value // 1024) / 1024)
		str(string("KB"))
	else
		dec(value)

PRI dec(value) | _i
'' Print a decimal number
	result := 0
	if value < 0
		chr("-")
		if (value == NEGX)
			chr("2")
			value += 2_000_000_000
		-value
	_i := 1_000_000_000
	repeat 10
		if value => _i
			chr(value / _i + "0")
			value //= _i
			result~~
		elseif result or _i == 1
			chr("0")
		_i /= 10

PRI dech(value) | _i
'' Print a decimal number with 1000s tick marks

	if value < 0
		chr("-")
		if (value == NEGX)
			chr("2")
			value += 2_000_000_000
		-value

	_i := 1_000_000_000
	repeat 10
		if value => _i
			if _i == 100_000_000 or _i == 100_000 or _i == 100
				chr("'")
			chr(value / _i + "0")
			value //= _i
			result~~
		elseif result or _i == 1
			if _i == 100_000_000 or _i == 100_000 or _i == 100
				chr("'")
			chr("0")
		_i /= 10

PRI hex(value, digits)
'' Print a hexadecimal number

	value <<= (8 - digits) << 2
	repeat digits
		chr(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PRI bin(value, digits)
'' Print a binary number

	value <<= 32 - digits
	repeat digits
		chr((value <-= 1) & 1 + "0")

PRI failure(who, val)
	str(who)
	str(string(" failed: "))
	dec(val)
	crlf

PRI success(who) | i, n, n1, n2, n3, n4, n5, n6, n7
	str(who)
	str(string(" initialized, "))
	dec(free_cogs)
	str(string(" cogs free.", 13, 10))

PUB free_cogs : free | i, cogs[8], jmp_0
	jmp_0 := %010111_0001_1111_000000000_000000000
	repeat while (cogs[free] := cognew(@jmp_0, 0)) => 0 
		free++
	if free
		repeat i from 0 to free - 1
			cogstop(cogs[i])
	return free

PRI direntry(n, buf) | rec, size, i
	rec := fat.opendir(@de)
	repeat while rec => 0
		if byte[@de][fat#dent_name] <> $e5	' deleted entry?
			byte[@de][fat#dent_attr] := 0
			if strcomp(@de + fat#dent_ext, string("CMD"))
				if n == 0
					fat.get_dent_name(@de, @name)
					i := strsize(@name)
					bytemove(buf, @name, i)
					byte[buf][i] := 0
					return 0
				n--
			rec := fat.readdir(rec, @de)
	return -1

PRI open_file(fname_ptr) : dent | size, sec
	dent := fat.find_file(fname_ptr)
	if dent < 0
		str(fname_ptr)
		str(string(" not found.",13,10))
		return -1
	bytemove(@de, dent, 32)
	fat.get_dent_name(dent, @name)
	size := fat.get_dent_size(@de)
	sec := fat.get_dent_first_sector(@de)
	fat.get_dent_date_time(@de, @datetime)
	str(@name)
	str(string(9,"sector "))
	dec(sec)
	str(string(9,"size "))
	dec_size(size)
	crlf

PRI cmd_load(fname_ptr) | dent, sec, nsec, pos, state, badr, blen, bmin, bmax
	dent := open_file(fname_ptr)
	if dent < 0
		return -1
	fat.get_dent_name(dent, @name)
	sec := fat.get_dent_first_sector(@de)

	pos := $200
	state := 0
	bmin := $ffff
	bmax := $0000
	repeat
		if pos == $200
			if sec < 0
				quit
			nsec := fat.get_dent_next_sector(@de, sec)
			fat.read_sector(sec)
			sec := nsec
			pos := 0
		case state
			0:	' expecting a header
				case byte[iobuff_ptr][pos++]
					$01:	' CMD load block header
						state := 1
					$02:	' CMD entry point header
						state := 7
					other:	' CMD comment block
						state := 5

			1:	' expecting a block length
				blen := byte[iobuff_ptr][pos++]
				state := 2

			2:	' expecting address LSB
				blen--
				badr := byte[iobuff_ptr][pos++]
				state := 3

			3:	' expecting address MSB
				blen--
				badr += 256 * byte[iobuff_ptr][pos++]
				state := 4
				if blen =< 0
					blen += 256
				bmin := bmin <# badr

			4:	' expecting blen bytes of data
				xmm.DoCmd("W", iobuff_ptr + pos, badr, 1)
				badr++
				pos++
				blen--
				if blen == 0
					bmax := bmax #> badr
					state := 0

			5:	' expecting a block length
				blen := byte[iobuff_ptr][pos++]
				if blen == 0
					blen := 256
				state := 6

			6:	' ignoring comment data
				pos++
				blen--
				if blen == 0
					state := 0

			7:	' expecting a block length
				blen := byte[iobuff_ptr][pos++]
				state := 8

			8:	' expecting entry point LSB
				blen--
				badr := byte[iobuff_ptr][pos++]
				state := 9

			9:	' expecting entry point MSB
				blen--
				badr += 256 * byte[iobuff_ptr][pos++]
				str(string("Loaded from "))
				hex(bmin, 4)
				str(string(".."))
				hex(bmax - 1, 4)
				crlf
				str(string("Entry point "))
				hex(badr, 4)
				crlf
				xmm.DoCmd("W", @badr, $0103, 2)
				quit

PUB command_addr
'' Return the address of the command long
	return @io_command

DAT
'---------------------------------------------------------------------------------------------------------

		org	0

'---------------------------------------------------------------------------------------------------------
entry
iocmd_ptr	mov	iocmd_ptr, PAR
		jmp	#run

screen_ptr	long	0-0
addr		long	0
cur_x		long	0
cur_y		long	0
iobuff_ptr	long	0-0
x20202020	long	$20202020
x10_000		long	10_000

dira_mask	long	0-0
ctra_mode	long	0-0
frqa_0		long	$7850_0000				' 0.46V of 1V, ie. 46%
frqa_1		long	$0000_0000				' 0V
frqa_2		long	$d999_0000				' 0.85V of 1V, ie. 85%
frqa_3		long	$0000_0000				' 0V

xmm_command	long	0-0
xmm_hubaddr	long	0-0
xmm_ramaddr	long	0-0
xmm_blocklen	long	0-0
xmm_errx	long	0-0

kbd_states	long	0-0
kbd_change	long	0-0
kbd_index	long	0
kbd_shift	long	0
kbd_data	long	0

minus1		long	$ffffffff
low_24bits	long	$00ffffff
low_16bits	long	$0000ffff
msg_break_ptr	long	@@@msg_break

iocmd		long	0
port		long	0
data		long	0
null		long	0

t1		long	0
t2		long	0
t3		long	0
t4		long	0

strptr		long	0
hexval		long	0
hexcnt		long	0

port_ff		long	0

run
		wrlong	null, iocmd_ptr

		mov	DIRA, dira_mask			' enable the sound output
		mov	CTRA, ctra_mode			' set the counter mode to single ended DUTY mode
		mov	FRQA, frqa_0			' set the default output pulse width

command
		rdlong	iocmd, iocmd_ptr
		mov	t1, iocmd
		shr	t1, #24			WZ
	if_z	jmp	#command
		cmp	t1, #io_cmd_out		WZ
	if_z	jmp	#output
		cmp	t1, #io_cmd_in		WZ
	if_z	jmp	#input
		cmp	t1, #io_cmd_break	WZ
	if_z	jmp	#break
		cmp	t1, #io_cmd_kbd		WZ
	if_z	jmp	#key_matrix
command_0
		and	iocmd, low_24bits
		wrlong	iocmd, iocmd_ptr
		jmp	#command

'*******************************************
'
' Read i/o port and data to write to
'
output
		mov	port, iocmd			' get command word
		shr	port, #8			' extract port
		and	port, #$ff			' lower 8 bits
		mov	data, iocmd			' get command word
		and	data, #$ff			' extract data

		cmp	port, #$ff		WZ	' port = $FF?
	if_z	jmp	#output_ff

		jmp	#command_0


output_ff
		mov	port_ff, data			' save value
		mov	t1, data
		and	t1, #3
		add	t1, #frqa_0
		movs	:padr, t1
		mov	t1, data
:padr		mov	t2, 0-0				' get audio level
		mov	FRQA, t2			' write to FRQA
		' TODO: check the other bits of data
		jmp	#command_0

'*******************************************
'
' Read i/o port to read from
'
input
		andn	iocmd, #$ff			' clear the input data
		mov	port, iocmd			' get command word
		shr	port, #8
		and	port, #$ff

		cmp	port, #$ff		WZ	' port = $FF?
	if_z	jmp	#input_ff

		or	iocmd, #$ff			' return $ff for other ports
		jmp	#command_0

input_ff
		mov	data, port_ff
		andn	data, #%00010000		' clear input bit
		' TODO: sample a pin to read from cassette ?
		shl	data, #3
		and	data, #%11000000
		or	iocmd, data
		jmp	#command_0
		
'*******************************************
'
' Output a character to the screen
'
conout_chr
		mov	addr, cur_y			' y
		shl	addr, #6			' * 64
		add	addr, cur_x			' + x
		add	addr, screen_ptr		' + screen address

		cmp	data, #8		WZ	' backspace?
	if_z	jmp	#:bsp
		cmp	data, #9		WZ	' tab?
	if_z	jmp	#:tab
		cmp	data, #10		WZ	' line feed?
	if_z	jmp	#:lf
		cmp	data, #12		WZ	' clear screen?
	if_z	jmp	#:cls
		cmp	data, #13		WZ	' carriage return?
	if_z	jmp	#:cr
:out
		wrbyte	data, addr
		add	cur_x, #1
:wrap
		cmp	cur_x, #64		WZ
	if_nz	jmp	#:exit
		mov	cur_x, #0
:lf
		add	cur_y, #1
		cmp	cur_y, #16		WZ
	if_nz	jmp	#:exit
		call	#scroll
		mov	cur_y, #15
		jmp	#:exit

:bsp
		cmp	cur_x, #0		WZ, WC
	if_be	jmp	#:exit
		sub	cur_x, #1
		jmp	#:exit
:tab
		mov	data, #32
		wrbyte	data, addr
		add	addr, #1
		add	cur_x, #1
		test	cur_x, #7		WZ
	if_nz	jmp	#:tab
		jmp	#:wrap
:cr
		mov	cur_x, #0
		jmp	#:exit
:cls
		mov	addr, screen_ptr
		mov	cur_y, #64*16/4
:loop
		wrlong	x20202020, addr
		add	addr, #4
		djnz	cur_y, #:loop
		mov	cur_x, #0
:exit
conout_chr_ret
		ret		

'*******************************************
'
' Scroll the screen one line up
'
scroll
		mov	addr, screen_ptr
		mov	cur_y, #15
:row
		mov	cur_x, #64
:col
		add	addr, #64
		rdlong	data, addr
		sub	addr, #64
		wrlong	data, addr
		add	addr, #4
		djnz	cur_x, #:col
		djnz	cur_y, #:row
		mov	cur_x, #64/4
:fill
		wrlong	x20202020, addr
		add	addr, #4
		djnz	cur_x, #:fill
scroll_ret
		ret


'*******************************************
'
' Output a byte in hexval to the console
'
conout_hex_b
		mov	hexcnt, #2
		shl	hexval, #24
		jmp	#conout_hex

'*******************************************
'
' Output a word in hexval to the console
'
conout_hex_w
		mov	hexcnt, #4
		shl	hexval, #16
		jmp	#conout_hex

'*******************************************
'
' Output a long in hexval to the console
'
conout_hex_l
		mov	hexcnt, #8
conout_hex
		mov	data, hexval
		shr	data, #28
		cmp	data, #10	WZ, WC
	if_b	add	data, #48
	if_ae	add	data, #55
		call	#conout_chr
		shl	hexval, #4
		djnz	hexcnt, #conout_hex
conout_hex_ret
conout_hex_l_ret
conout_hex_w_ret
conout_hex_b_ret
		ret

'*******************************************
'
' Output a carriage return to the console
'
conout_space
		mov	data, #32
		call	#conout_chr
conout_space_ret
		ret

'*******************************************
'
' Output a carriage return to the console
'
conout_crlf
		mov	data, #13
		call	#conout_chr
		mov	data, #10
		call	#conout_chr
conout_crlf_ret
		ret

'*******************************************
'
' Output a string at strptr to the console
'
conout_string
		rdbyte	data, strptr	WZ
		add	strptr, #1
	if_z	jmp	#conout_string_ret
		call	#conout_chr
		jmp	#conout_string
conout_string_ret
		ret

'*******************************************
'
' Break - issued when the Z80 encounters
' a HALT instruction.
'
break
		mov	t1, iocmd			' read the address of the register file
		shr	t1, #8				' bits 23..8 to 15..0
		and	t1, low_16bits			' mask only bits 15..0
		mov	strptr, msg_break_ptr	

		mov	t2, #8
:loop_r16	call	#conout_string
		rdword	hexval, t1			' read BC, DE, HL, AF, BC', DE', HL', AF'
		call	#conout_hex_w
		add	t1, #2
		djnz	t2, #:loop_r16

		call	#conout_string			' new line and print "R:"
		rdbyte	hexval, t1			' R
		add	t1, #1
		rdbyte	t2, t1				' R2
		and	hexval, #%01111111		' R register's bit #7
		or	hexval, t2			' does never change from the last written value
		call	#conout_hex_b
		add	t1, #1

		call	#conout_string
		rdbyte	hexval, t1			' IFF1+2/IM
		mov	t2, t1				' save value
		and	hexval, #1			' IFF1 (interrupt flip-flop 1)
		call	#conout_hex_b

		call	#conout_string
		mov	hexval, t2			' IFF1+2/IM
		shr	hexval, #1			' shift bit #1 to #0
		and	hexval, #1			' IFF2 (interrupt flip-flop 2)
		call	#conout_hex_b

		call	#conout_string
		mov	hexval, t2			' IFF1+2/IM
		shr	hexval, #2			' shift bits #3..2 to #1..0
		and	hexval, #3			' IM (interrupt mode)
		call	#conout_hex_b
		add	t1, #1

		call	#conout_string
		rdbyte	hexval, t1			' I (interrupt vector register)
		call	#conout_hex_b
		add	t1, #1

		call	#conout_string
		rdword	hexval, t1			' IX
		call	#conout_hex_w
		add	t1, #2

		call	#conout_string
		rdword	hexval, t1			' IY
		call	#conout_hex_w
		add	t1, #2

		call	#conout_string
		rdword	hexval, t1			' SP
		call	#conout_hex_w
		add	t1, #2

		call	#conout_string
		rdword	hexval, t1			' PC
		mov	t3, hexval
		call	#conout_hex_w
		call	#conout_string
		add	t1, #2				' skip to opcode

		wrlong	t3, xmm_ramaddr			' as XMM ram address
		mov	t2, #4				' 4 bytes
		wrlong	t2, xmm_blocklen		' blocklen to read
		mov	t2, #"R"			' XMM read command
		wrlong	t2, xmm_command
':wait		rdlong	t2, xmm_command		WZ
'	if_nz	jmp	#:wait

		rdbyte	hexval, t1
		call	#conout_hex_b

		call	#conout_crlf

		jmp	#command_0

'*******************************************
'
' Read the TRS-80 keyboard matrix
' Address is in io_port, address bits 0..7
' determine the keyboard row(s) to read
'
key_matrix
		mov	port, iocmd			' get command word
		shr	port, #8
		and	port, low_16bits
		mov	t1, #0				' resulting bits
		mov	t2, kbd_states			' address of the keyboard state matrix
		mov	t3, #8				' 8 of them
		mov	data, #$01			' bit mask for the address lines
:loop		rdbyte	kbd_data, t2			' next keyboard row entry
		add	t2, #1
		test	port, data		WZ	' test port address A0, A1, A2 ...
	if_nz	or	t1, kbd_data			' if address bit was set, OR the data
		rol	data, #1			' rotate mask left
		djnz	t3, #:loop			' 8 times
		or	iocmd, t1			' return it in the command word bits 7..0
		jmp	#command_0

		fit	$1f0

'***************************************************************************
'*   w/o SHIFT                             with SHIFT
'*   +-------------------------------+     +-------------------------------+
'*   | 0   1   2   3   4   5   6   7 |     | 0   1   2   3   4   5   6   7 |
'*+--+---+---+---+---+---+---+---+---+  +--+---+---+---+---+---+---+---+---+
'*|0 | @ | A | B | C | D | E | F | G |  |0 | ` | a | b | c | d | e | f | g |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|1 | H | I | J | K | L | M | N | O |  |1 | h | i | j | k | l | m | n | o |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|2 | P | Q | R | S | T | U | V | W |  |2 | p | q | r | s | t | u | v | w |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|3 | X | Y | Z | [ | \ | ] | ^ | _ |  |3 | x | y | z | { | | | } | ~ |   |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|4 | 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 |  |4 |   | ! | " | # | $ | % | & | ' |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|5 | 8 | 9 | : | ; | , | - | . | / |  |5 | ( | ) | * | + | < | = | > | ? |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|6 |ENT|CLR|BRK|UP |DN |LFT|RGT|SPC|  |6 |ENT|CLR|BRK|UP |DN |LFT|RGT|SPC|
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|7 |SHF|ALT|PUP|PDN|INS|DEL|CTL|END|  |7 |SHF|ALT|PUP|PDN|INS|DEL|CTL|END|
'*+--+---+---+---+---+---+---+---+---+  +--+---+---+---+---+---+---+---+---+
'***************************************************************************

		align	long
kbd_rows_table
		byte	0,0,0,0,0,0,0,0

msg_break	byte	"BC:",0
		byte	" DE:",0
		byte	" HL:",0
		byte	" AF:",0
		byte	" BC'",0
		byte	" DE'",0
		byte	" HL'",0
		byte	" AF'",0
		byte	13,10, "R:",0
		byte	" IFF1:",0
		byte	" IFF2:",0
		byte	" IM:",0
		byte	" I:",0
		byte	" IX:",0
		byte	" IY:",0
		byte	" SP:",0
		byte	" PC:",0
		byte	" = ",0

''      MIT LICENSE
{{
'  Permission is hereby granted, free of charge, to any person obtaining
'  a copy of this software and associated documentation files
'  (the "Software"), to deal in the Software without restriction,
'  including without limitation the rights to use, copy, modify, merge,
'  publish, distribute, sublicense, and/or sell copies of the Software,
'  and to permit persons to whom the Software is furnished to do so,
'  subject to the following conditions:
'
'  The above copyright notice and this permission notice shall be included
'  in all copies or substantial portions of the Software.
'
'  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}}
